import re

from .exceptions import SNSparserException

class SNSinteraction:

    def __init__(self, type, action, headers={}, content=None):
        self.type = str(type)
        self.action = str(action)
        self.headers = headers
        self.content = content

        if self.content is not None:
            self.headers['Content-Length'] = str(len(self.content))
        else:
            self.headers['Content-Length'] = '0'

    def __str__(self):
        interaction_string = self.type + " " + self.action

        if len(self.headers) > 0:
            interaction_string += "\n"
            interaction_string += "\n".join([header_field[0] + ": " + str(header_field[1]) for header_field in self.headers.items()])

        if not self.content is None and not len(self.content) == 0:
            interaction_string += "\n\n"
            interaction_string += self.content

        return interaction_string

def parse(interaction):
    """
    Parses a string representing a interaction made with the SNS protocol.
    """

    interaction_lines = interaction.split('\n\n')
    headers = interaction_lines[0].splitlines()

    if len(interaction_lines) > 2:
        raise SNSparserException(interaction, 'The interaction has two empty lines, while only one after the headers is allowed')

    try:
        interaction_method = headers[0]
    except IndexError:
        raise SNSparserException(interaction, 'The interaction method cannot be parsed, missing the first line')

    try:
        _interaction_type, _interaction_action = interaction_method.split(" ")

        if len(_interaction_type) == 0:
            raise SNSparserException(interaction_method, 'The interaction type is empty')

        if not _interaction_type.isupper():
            raise SNSparserException(interaction_method, 'The interaction type is not capitalized, suggesting it is not a proper interaction type')

        if len(_interaction_action) == 0:
            raise SNSparserException(interaction_method, 'The interaction action is empty')

        if not _interaction_action.islower() and re.match(r'^-?\d+(?:\.\d+)?$', _interaction_action) is None:
            raise SNSparserException(interaction_method, 'The interaction action contains capitalized letters, suggesting it is not a proper interaction action')
    except ValueError:
        raise SNSparserException(interaction_method, 'The interaction method does not contain the interaction type and action')

    header_fields = headers[1:]
    _interaction_headers = {}
    
    for header_field in header_fields:
        try:
            header_type, header_value = header_field.split(': ')

            if len(header_type) == 0:
                raise SNSparserException(header_field, 'The header type is empty')

            if len(header_value) == 0:
                raise SNSparserException(header_field, 'The header value is empty')

            if header_type in _interaction_headers.keys():
                raise SNSparserException(header_field, 'This header type is a duplicate and is already set')

            _interaction_headers[header_type] = header_value
        except ValueError:
            raise SNSparserException(header_field, 'The header field does not contain a type and value')

    if "From" not in _interaction_headers.keys():
        raise SNSparserException(headers, 'The interaction is missing the From field, make sure this is set as a header')

    if not len(_interaction_headers['From'].split(":")) == 2:
        raise SNSparserException(headers, 'The identifier in the From header is missing an interaction qualifier, e.g. \'user\' or \'group\'')

    if "To" not in _interaction_headers.keys():
        raise SNSparserException(headers, 'The interaction is missing the To field, make sure this is set as a header')

    if not len(_interaction_headers['To'].split(":")) == 2:
        raise SNSparserException(headers, 'The identifier in the To header is missing an interaction qualifier, e.g. \'user\' or \'group\'')

    _interaction_content = None
    if len(interaction_lines) == 2:
        _interaction_content = interaction_lines[1]

    return SNSinteraction(_interaction_type, _interaction_action, _interaction_headers, _interaction_content)


from .parser import SNSinteraction, parse
from .exceptions import SNSinteractionException
from .records import SNSuserRecord

import socket

SNS_PORT = 3735

#TODO: encrypt the whole request with the private key of the user first, to ensure integrity and authenticity. 
#Since the public key is already known to the server, this does not require any extra steps either.
def request(type, action, content=None, **headers):
    for key, value in headers.items():
        try:
            headers[key] = str(value)
        except IndexError:
            raise SNSinteractionException(str(headers), "One of the interaction headers could not be parsed to a string")

    interaction = parse(str(SNSinteraction(type, action, headers, content)))

    if interaction.headers['To'].count('@') > 1:
        raise SNSinteractionException(str(interaction), "The To header contains too many @ characters (max. 1 per address header), causing that the address cannot be determined")

    try:
        address = interaction.headers['To'].split('@')[1]
    except:
        address = SNSuserRecord.provider

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((address, SNS_PORT))
    s.sendall(str(interaction).encode())
    data = s.recv(4096)
    s.close()

    interaction_strings = data.decode("utf-8").split("\n\n")
    interaction_strings = ["\n\n".join(interaction_strings[i:i+2]) for i in range(0, len(interaction_strings), 2)]
    
    return [parse(interaction_string) for interaction_string in interaction_strings]

def user(action, To, content=None, **headers):
    encrypted_content = content

    headers['From'] = 'user:' + SNSuserRecord.identifier
    headers['To'] = To

    interaction_object = request('USER', action, encrypted_content, **headers)
    
    return interaction_object

def group(action, To, content=None, **headers):
    encrypted_content = content

    headers['From'] = 'user:' + SNSuserRecord.identifier
    headers['To'] = To

    interaction_object = request('GROUP', action, encrypted_content, **headers)
    
    return interaction_object

def post(action, To, content=None, **headers):
    encrypted_content = content

    headers['From'] = 'user:' + SNSuserRecord.identifier
    headers['To'] = To

    interaction_object = request('POST', action, encrypted_content, **headers)
    
    return interaction_object

def reaction(action, To, content=None, **headers):
    encrypted_content = content

    headers['From'] = 'user:' + SNSuserRecord.identifier
    headers['To'] = To

    interaction_object = request('REACTION', action, encrypted_content, **headers)
    
    return interaction_object


import socket
import threading

from .parser import parse, SNSinteraction

class SNSserver(object):

    HOST = '0.0.0.0'
    PORT = 3735

    components = {}

    server_thread = None
    conn_threads = []

    def __init__(self, type=None, action=None):
        self.type = type
        self.action = action

        if self.type is None:
            self.type = "*"

        if self.action is None:
            self.action = "*"

    def __call__(self, func):
        if not isinstance(self.type, list):
            self.type = [self.type]

        if not isinstance(self.action, list):
            self.action = [self.action]

        for type in self.type:
            for action in self.action:
                self.components[type, action] = func

    @staticmethod
    def start_server():
        SNSserver.server_thread = threading.Thread(target=SNSserver.server_thread)
        SNSserver.server_thread.start()

    @staticmethod
    def server_thread():
        print("Starting on:", SNSserver.HOST, SNSserver.PORT)
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind((SNSserver.HOST, SNSserver.PORT))
        s.listen(1)

        print("Server has started..")

        while True:
            conn, addr = s.accept()

            conn_thread = threading.Thread(target=SNSserver.connection_handler, args=(conn, addr))
            conn_thread.start()
            
            SNSserver.conn_threads.append(conn_thread)

        s.close()

    @staticmethod
    def connection_handler(conn, addr):
        with conn:
            print('Connected with', addr)
            conn.settimeout(2)
            while True:
                try:
                    data = conn.recv(4096)

                    if not data:
                        break

                    interaction_object = parse(data.decode("utf-8"))

                    print(addr, "Received interaction:")
                    print(interaction_object)

                    if (interaction_object.type, interaction_object.action) in SNSserver.components:
                        response = SNSserver.components[interaction_object.type, interaction_object.action](interaction_object)
                    elif (interaction_object.type, "*") in SNSserver.components:
                        response = SNSserver.components[interaction_object.type, "*"](interaction_object)
                    elif ("*", interaction_object.action) in SNSserver.components:
                        response = SNSserver.components["*", interaction_object.action](interaction_object)
                    elif ("*", "*") in SNSserver.components:
                        response = SNSserver.components["*", "*"](interaction_object)
                    else:
                        print(addr, "Interaction component not found")
                        response = SNSinteraction("STATUS", 404, {'From':interaction_object.headers['To'], 'To':interaction_object.headers['From']})

                    print(addr, "Constructed response:")
                    if isinstance(response, list):
                        if len(response) == 0:
                            response = SNSinteraction("STATUS", 401, {'From':interaction_object.headers['To'], 'To':interaction_object.headers['From']})
                            print(response)
                            conn.sendall(str(response).encode("utf-8"))
                        else:
                            print(response)
                            conn.sendall('\n\n'.join([str(item) for item in response]).encode("utf-8"))
                    else:
                        print(response)
                        conn.sendall(str(response).encode("utf-8"))
                except socket.error:
                    break
        conn.close()


import unittest

from .parser import parse, SNSinteraction
from .exceptions import SNSparserException

class TestSNSParser(unittest.TestCase):

    def test_parse_user_follow(self):
        interaction = """USER follow
From: user:alice@snsprovider.net
To: user:bob@snsservice.com"""
        
        interaction_object = parse(interaction)
        assert not interaction_object == None
        assert interaction_object.type == "USER"
        assert interaction_object.action == "follow"
        assert interaction_object.headers == {'From':'user:alice@snsprovider.net','To':'user:bob@snsservice.com'}
        assert interaction_object.content is None

    def test_parse_empty(self):
        self.assertRaisesRegex(SNSparserException, 
        'The interaction method cannot be parsed, missing the first line', parse, '')

    def test_parse_missing_type(self):
        self.assertRaisesRegex(SNSparserException, 
        '.*: The interaction type is empty', parse, ' follow')

    def test_parse_missing_action(self):
        self.assertRaisesRegex(SNSparserException, 
        '.*: The interaction action is empty', parse, 'USER ')

    def test_parse_lower_type(self):
        self.assertRaisesRegex(SNSparserException, 
        '.*: The interaction type is not capitalized, suggesting it is not a proper interaction type', parse, 'uSeR follow')

    def test_parse_upper_action(self):
        self.assertRaisesRegex(SNSparserException, 
        '.*: The interaction action contains capitalized letters, suggesting it is not a proper interaction action', parse, 'USER Follow')

    def test_parse_with_content(self):
        interaction = """POST create
From: user:alice@snsprovider.net
To: user:bob@snsservice.com

Hello"""
        
        interaction_object = parse(interaction)
        assert not interaction_object == None
        assert interaction_object.type == "POST"
        assert interaction_object.action == "create"
        assert interaction_object.headers == {'From':'user:alice@snsprovider.net','To':'user:bob@snsservice.com'}
        assert interaction_object.content == "Hello"

    def test_double_parse(self):
        interaction = """POST create
From: user:alice@snsprovider.net
To: user:bob@snsservice.com
Limit: 3
Do: 5

Hello"""
        
        interaction_object = parse(interaction)
        assert not interaction_object == None
        assert interaction_object.type == "POST"
        assert interaction_object.action == "create"
        assert interaction_object.headers == {'From':'user:alice@snsprovider.net','To':'user:bob@snsservice.com', 'Limit':'3', 'Do':'5'}
        assert interaction_object.content == "Hello"
        
        interaction_object = parse(str(interaction_object))
        assert not interaction_object == None
        assert interaction_object.type == "POST"
        assert interaction_object.action == "create"
        assert interaction_object.headers == {'From':'user:alice@snsprovider.net','To':'user:bob@snsservice.com', 'Limit':'3', 'Do':'5'}
        assert interaction_object.content == "Hello"

    def test_parse_content_no_break(self):
        self.assertRaisesRegex(SNSparserException, 
        '.*: The header field does not contain a type and value', parse, """POST create
From: alice@snsprovider.net
To: bob@snsservice.com
Hello""")

    def test_content_two_breaks(self):
        self.assertRaisesRegex(SNSparserException, 
        '.*: The interaction has two empty lines, while only one after the headers is allowed', parse, """POST create
From: alice@snsprovider.net
To: bob@snsservice.com

Hello

Hello again""")

    def test_missing_type_or_action(self):
        self.assertRaisesRegex(SNSparserException, 
        '.*: The interaction method does not contain the interaction type and action', parse, 'USER')

    def test_too_many_types_or_actions(self):
        self.assertRaisesRegex(SNSparserException, 
        '.*: The interaction method does not contain the interaction type and action', parse, 'USER follow unfollow')

    def test_missing_header_type(self):
        self.assertRaisesRegex(SNSparserException, 
        '.*: The header type is empty', parse, """USER follow
: alice?""")

    def test_missing_header_value(self):
        self.assertRaisesRegex(SNSparserException, 
        '.*: The header value is empty', parse, """USER follow
To: """)

    def test_invalid_header_field(self):
        self.assertRaisesRegex(SNSparserException, 
        '.*: The header field does not contain a type and value', parse, """USER follow
To:ss""")

    def test_too_many_header_items(self):
        self.assertRaisesRegex(SNSparserException, 
        '.*: The header field does not contain a type and value', parse, """USER follow
To: alice: or bob""")

    def test_missing_from_header(self):
        self.assertRaisesRegex(SNSparserException, 
        '.*: The interaction is missing the From field, make sure this is set as a header', parse, """USER follow
To: alice""")

    def test_missing_to_header(self):
        self.assertRaisesRegex(SNSparserException, 
        '.*: The interaction is missing the To field, make sure this is set as a header', parse, """USER follow
From: user:alice""")

    def test_header_duplicate(self):
        self.assertRaisesRegex(SNSparserException, 
        '.*: This header type is a duplicate and is already set', parse, """USER follow
From: alice
From: Bob""")

    def test_from_header_missing_qualifier(self):
        self.assertRaisesRegex(SNSparserException, 
        '.*: The identifier in the From header is missing an interaction qualifier, e.g. \'user\' or \'group\'', parse, """USER follow
From: alice
To: user:Bob""")

    def test_to_header_missing_qualifier(self):
        self.assertRaisesRegex(SNSparserException, 
        '.*: The identifier in the To header is missing an interaction qualifier, e.g. \'user\' or \'group\'', parse, """USER follow
From: user:alice
To: Bob""")

if __name__ == '__main__':
    unittest.main()


from .server import SNSserver
from .parser import SNSinteraction

SNSserver.start_server()

@SNSserver(type="USER", action="*")
def follow_user(request):
    return SNSinteraction("STATUS", 200, {})

@SNSserver(type="GROUP", action="join")
def join_group(request):
    return SNSinteraction("STATUS", 201, {})


class SNSparserException(Exception):
    """Exception raised when parsing an SNS interaction.

    Attributes:
        interaction -- the interaction that raised the error
        message -- explanation of the error
    """

    def __init__(self, interaction, message="Exception occurred while parsing the following interaction"):
        self.interaction = interaction
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.interaction}: {self.message}'

class SNSinteractionException(Exception):
    """Exception raised when requesting an SNS interaction.
    
    Attributes:
        interaction -- the interaction that raised the error
        message -- explanation of the error
    """

    def __init__(self, interaction, message="Exception occurred while requesting with the following interaction"):
        self.interaction = interaction
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.interaction}: {self.message}'

class SNSrecordIntegrityException(Exception):
    """Exception raised when the integrity of a record is instable.
    
    Attributes:
        record -- the record that raised the error
        message -- explanation of the error
    """

    def __init__(self, record, message="Exception occurred while checked the integrity of the following record"):
        self.record = record
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.record}: {self.message}'


from os.path import expanduser, isdir, isfile
from os import mkdir, mknod

from .exceptions import SNSrecordIntegrityException

SNS_RECORD_PATH = "~/.sns_records"

SNSuserRecord = None

class SNSuserRecordInstance():
    def __init__(self, identifier, provider):
        self.identifier = identifier
        self.provider = provider

def register_sns_records():
    global SNSuserRecord
    
    check_record_integrity(restore=True)

    with open(expanduser(SNS_RECORD_PATH) + "/user_record.dat") as user_record:
        record_items = dict(item.split(":", 1) for item in user_record.read().split(";") if ':' in item)

        SNSuserRecord = SNSuserRecordInstance(record_items['identifier'], record_items['provider'])
    
def check_record_integrity(restore=False):
    if not isdir(expanduser(SNS_RECORD_PATH)):
        if restore:
            mkdir(expanduser(SNS_RECORD_PATH))
        else:
            raise SNSrecordIntegrityException("User Record", "Missing the SNS records folder. Set the restore flag to True to restore the folder")

    if not isfile(expanduser(SNS_RECORD_PATH) + "/user_record.dat"):
        if restore:
            user_record = open(expanduser(SNS_RECORD_PATH) + "/user_record.dat", "w+")
            user_record.write("identifier:you;")
            user_record.write("provider:192.168.178.50")
            user_record.close()
        else:
            raise SNSrecordIntegrityException("User Record", "Missing the UserRecord file. Set the restore flag to True to restore the file.\n[WARNING] This will create new key pairs")

register_sns_records()